from .pai_requests import PAIGetRequests, PAIPostRequests
from .pai_responses import (
    AnalyzeTextResponse,
    BleepResponse,
    DiagnosticResponse,
    FilesBase64Response,
    FilesUriResponse,
    MetricsResponse,
    NerTextResponse,
    ReidentifyTextResponse,
    TextResponse,
    VersionResponse,
)
from .pai_uris import PAIURIs
from .request_objects import *



import base64
import os

from dotenv import load_dotenv

from ..objects import request_objects as rq
from ..pai_client import PAIClient

load_dotenv()


def _get_client():
    client = PAIClient(url=os.environ["PAI_API_URL"])
    if os.getenv("PAI_API_KEY"):
        client.add_api_key(os.getenv("PAI_API_KEY"))
    return client


def test_ping():
    client = _get_client()
    assert client.ping() == True


def test_text_only():
    client = _get_client()
    req = rq.process_text_obj(text=["Hey there!"])
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_default():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj()
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_coref_model_prediction():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        processed_text=rq.processed_text_obj(
            type="MARKER", coreference_resolution="model_prediction"
        ),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_coref_heuristic():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        processed_text=rq.processed_text_obj(
            type="MARKER", coreference_resolution="heuristics"
        ),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_coref_combined():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        processed_text=rq.processed_text_obj(
            type="MARKER", coreference_resolution="combined"
        ),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_with_enable_entity_types():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.process_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(entity_types=[selector]),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_with_disable_entity_types():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="DISABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.process_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(entity_types=[selector]),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_with_allow_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    req = rq.process_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj(filter=[filter])
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_with_block_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(
        type="BLOCK", pattern="[A-Za-z0-9]*", entity_type="ANY_TEXT"
    )
    req = rq.process_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj(filter=[filter])
    )
    resp = client.process_text(req)
    assert resp.ok


def test_entity_detection_with_allow_text_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW_TEXT", pattern="[A-Za-z0-9]*")
    req = rq.process_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj(filter=[filter])
    )
    resp = client.process_text(req)
    assert resp.ok


def test_full_entity_detection():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.process_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(
            filter=[filter], entity_types=[selector]
        ),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_processed_text_default():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"], processed_text=rq.processed_text_obj()
    )
    resp = client.process_text(req)
    assert resp.ok


def test_processed_text_marker():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"], processed_text=rq.processed_text_obj(type="MARKER")
    )
    resp = client.process_text(req)
    assert resp.ok


def test_processed_text_marker_language():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"],
        processed_text=rq.processed_text_obj(type="MARKER", marker_language="de"),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_processed_text_mask():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"], processed_text=rq.processed_text_obj(type="MASK")
    )
    resp = client.process_text(req)
    assert resp.ok


def test_processed_text_synthetic():
    client = _get_client()
    req = rq.process_text_obj(
        text=["Hey there!"], processed_text=rq.processed_text_obj(type="SYNTHETIC")
    )
    resp = client.process_text(req)
    assert resp.ok


def test_project_id():
    client = _get_client()
    req = rq.process_text_obj(text=["Hey there!"], project_id="test")
    resp = client.process_text(req)
    assert resp.ok


def test_full_text_request():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    req = rq.process_text_obj(
        text=["Hey there!"],
        processed_text=rq.processed_text_obj(type="MARKER"),
        entity_detection=rq.entity_detection_obj(
            filter=[filter], entity_types=[selector]
        ),
    )
    resp = client.process_text(req)
    assert resp.ok


def test_reidentify_text():
    client = _get_client()
    reid_req = rq.reidentify_text_obj(
        processed_text=["this", "is", "test", "data"],
        entities=[
            rq.entity(processed_text="this", text="that"),
            rq.entity(processed_text="is", text="was"),
            rq.entity(processed_text="test", text="real"),
            rq.entity(processed_text="data", text="nonsense"),
        ],
    )
    resp = client.reidentify_text(reid_req)
    assert resp.ok


def test_process_file_base64():
    client = _get_client()
    test_dir = "/".join(__file__.split("/")[:-1])
    with open(f"{test_dir}/test_files/simpsons_wiki.txt", "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")
    file_obj = rq.file_obj(data=file_data, content_type="text/plain")
    request_obj = rq.file_base64_obj(file=file_obj)
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_process_pptx_file_base64():
    client = _get_client()
    test_dir = "/".join(__file__.split("/")[:-1])
    with open(f"{test_dir}/test_files/demo-simple.pptx", "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")
    file_obj = rq.file_obj(
        data=file_data,
        content_type="application/vnd.openxmlformats-officedocument.presentationml.presentation",
    )
    request_obj = rq.file_base64_obj(file=file_obj)
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_process_dcm_file_base64():
    client = _get_client()
    test_dir = "/".join(__file__.split("/")[:-1])
    with open(f"{test_dir}/test_files/demo-sequence.dcm", "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")
    file_obj = rq.file_obj(data=file_data, content_type="application/dicom")
    request_obj = rq.file_base64_obj(file=file_obj)
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_process_audio_file_base64():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_audio.mp3"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "audio/mp3"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    audio_option_obj = rq.audio_options_obj(bleep_gain=-50, bleep_frequency=300)
    request_obj = rq.file_base64_obj(file=file_obj, audio_options=audio_option_obj)
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_process_image_file_base64():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_image.jpg"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "image/jpg"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    image_option_obj = rq.image_options_obj(masking_method="blur", palette=True)
    request_obj = rq.file_base64_obj(file=file_obj, image_options=image_option_obj)
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_process_ocr_image_file_base64():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_image.jpg"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "image/jpg"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    image_option_obj = rq.image_options_obj(masking_method="blur", palette=True)
    ocr_option_obj = rq.ocr_options_obj(ocr_system="azure_doc_intelligence")
    request_obj = rq.file_base64_obj(
        file=file_obj, image_options=image_option_obj, ocr_options=ocr_option_obj
    )
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_object_entity_detection_default():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_image.jpg"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "image/jpg"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    request_obj = rq.file_base64_obj(
        file=file_obj, object_entity_detection=rq.object_entity_detection_obj()
    )
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_object_entity_detection_with_enable_entity_types():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_image.jpg"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "image/jpg"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    selector = rq.object_entity_type_selector_obj(type="ENABLE", value=["LOGO"])
    request_obj = rq.file_base64_obj(
        file=file_obj,
        object_entity_detection=rq.object_entity_detection_obj(
            object_entity_types=[selector]
        ),
    )
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_object_entity_detection_with_disable_entity_types():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_image.jpg"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "image/jpg"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    selector = rq.object_entity_type_selector_obj(type="DISABLE", value=["SIGNATURE"])
    request_obj = rq.file_base64_obj(
        file=file_obj,
        object_entity_detection=rq.object_entity_detection_obj(
            object_entity_types=[selector]
        ),
    )
    resp = client.process_files_base64(request_object=request_obj)
    assert resp.ok


def test_bleep():
    client = _get_client()

    test_dir = "/".join(__file__.split("/")[:-1])
    file_name = "test_audio.mp3"
    filepath = os.path.join(f"{test_dir}", "test_files", file_name)
    file_type = "audio/mp3"

    with open(filepath, "rb") as b64_file:
        file_data = base64.b64encode(b64_file.read())
        file_data = file_data.decode("ascii")

    file_obj = rq.file_obj(data=file_data, content_type=file_type)
    timestamp = rq.timestamp_obj(start=1.0, end=2.0)

    request_obj = rq.bleep_obj(
        file=file_obj, timestamps=[timestamp], bleep_frequency=500, bleep_gain=-30
    )
    resp = client.bleep(request_object=request_obj)
    assert resp.ok


def test_ner_text_only():
    client = _get_client()
    req = rq.ner_text_obj(text=["Hey there!"])
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_entity_detection_default():
    client = _get_client()
    req = rq.ner_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj()
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_entity_detection_with_enable_entity_types():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.ner_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(entity_types=[selector]),
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_entity_detection_with_disable_entity_types():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="DISABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.ner_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(entity_types=[selector]),
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_entity_detection_with_allow_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    req = rq.ner_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj(filter=[filter])
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_entity_detection_with_block_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(
        type="BLOCK", pattern="[A-Za-z0-9]*", entity_type="ANY_TEXT"
    )
    req = rq.ner_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj(filter=[filter])
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_entity_detection_with_allow_text_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW_TEXT", pattern="[A-Za-z0-9]*")
    req = rq.ner_text_obj(
        text=["Hey there!"], entity_detection=rq.entity_detection_obj(filter=[filter])
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_ner_full_entity_detection():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.ner_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(
            filter=[filter], entity_types=[selector]
        ),
    )
    resp = client.ner_text(req)
    assert resp.ok


def test_analyze_text_only():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_entity_detection_default():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"], locale="en", entity_detection=rq.entity_detection_obj()
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_entity_detection_with_enable_entity_types():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
        entity_detection=rq.entity_detection_obj(entity_types=[selector]),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_entity_detection_with_disable_entity_types():
    client = _get_client()
    selector = rq.entity_type_selector_obj(type="DISABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
        entity_detection=rq.entity_detection_obj(entity_types=[selector]),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_entity_detection_with_allow_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
        entity_detection=rq.entity_detection_obj(filter=[filter]),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_entity_detection_with_block_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(
        type="BLOCK", pattern="[A-Za-z0-9]*", entity_type="ANY_TEXT"
    )
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
        entity_detection=rq.entity_detection_obj(filter=[filter]),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_entity_detection_with_allow_text_filter():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW_TEXT", pattern="[A-Za-z0-9]*")
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
        entity_detection=rq.entity_detection_obj(filter=[filter]),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_full_entity_detection():
    client = _get_client()
    filter = rq.filter_selector_obj(type="ALLOW", pattern="[A-Za-z0-9]*")
    selector = rq.entity_type_selector_obj(type="ENABLE", value=["HIPAA_SAFE_HARBOR"])
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        locale="en",
        entity_detection=rq.entity_detection_obj(
            filter=[filter], entity_types=[selector]
        ),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_coref_model_prediction():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        relation_detection=rq.relation_detection_obj(
            coreference_resolution="model_prediction"
        ),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_coref_heuristic():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        relation_detection=rq.relation_detection_obj(
            coreference_resolution="heuristics"
        ),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_coref_combined():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        relation_detection=rq.relation_detection_obj(coreference_resolution="combined"),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_enable_relation_extraction_true():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        relation_detection=rq.relation_detection_obj(
            coreference_resolution="combined", enable_relation_extraction=True
        ),
    )
    resp = client.analyze_text(req)
    assert resp.ok


def test_analyze_enable_relation_extraction_false():
    client = _get_client()
    req = rq.analyze_text_obj(
        text=["Hey there!"],
        entity_detection=rq.entity_detection_obj(),
        relation_detection=rq.relation_detection_obj(
            coreference_resolution="combined", enable_relation_extraction=False
        ),
    )
    resp = client.analyze_text(req)
    assert resp.ok



from collections import defaultdict
from typing import Literal

from pyxdameraulevenshtein import damerau_levenshtein_distance


class FuzzyMatchEntityProcessor:
    def __init__(
        self,
        known_words_list: list[str] | tuple[str] | set[str],
        threshold: int,
        strategy: Literal["BLOCK", "ALLOW"] = "BLOCK",
        process_type: Literal["MARKER", "MASK"] = "MARKER",
        masking_character: str = "#",
        ignore_casing: bool = True,
    ):
        self.known_words_list = known_words_list
        self.threshold = threshold
        self.process_type = process_type
        self.strategy = strategy
        self.ignore_casing = ignore_casing
        self.masking_character = masking_character
        self.counts: defaultdict[str, int] = defaultdict(int)
        self._validate_attributes()

    def __call__(self, entity: dict) -> str:
        if self.ignore_casing:
            lower_text = entity["text"].lower()
            min_dist = min(
                damerau_levenshtein_distance(lower_text, word.lower())
                for word in self.known_words_list
            )
        else:
            min_dist = min(
                damerau_levenshtein_distance(entity["text"], word)
                for word in self.known_words_list
            )
        should_allow = self.strategy == "ALLOW"
        is_similar = min_dist <= self.threshold

        if is_similar == should_allow:
            return entity["text"]
        else:
            if self.process_type == "MASK":
                return self.masking_character * len(entity["text"])
            key = entity["best_label"]
            self.counts[key] += 1
            return f"[{key}_{self.counts[key]}]"

    def _validate_attributes(self):
        if self.strategy not in ["BLOCK", "ALLOW"]:
            raise ValueError(
                f"Invalid value for strategy. Accepted values: 'BLOCK' and 'ALLOW'"
            )
        if self.process_type not in ["MARKER", "MASK"]:
            raise ValueError(
                f"Invalid value for process_type. Accepted values: 'MARKER' and 'MASK'"
            )
        if not isinstance(self.threshold, int) or self.threshold < 1:
            raise ValueError(
                f"Invalid value for threshold. Accepted value is a positive integer."
            )
        if not isinstance(self.known_words_list, (list, tuple, set)) or not all(
            isinstance(word, str) for word in self.known_words_list
        ):
            raise ValueError(
                f"Invalid value for known_words_list. Accepted are list, tuple or set of strings."
            )
        if not isinstance(self.masking_character, str):
            raise ValueError(
                f"Invalid value for masking_character. Accepted value is a valid string"
            )
        if not isinstance(self.ignore_casing, bool):
            raise ValueError(
                f"Invalid value for ignore_casing. Accepted values: True and False"
            )



from collections import defaultdict


class MaskEntityProcessor:
    def __init__(self, masking_character: str = "#"):
        self.masking_character = masking_character

    def __call__(self, entity: dict) -> str:
        return self.masking_character * len(entity["text"])


class MarkerEntityProcessor:
    def __init__(self):
        self.counts: defaultdict[str, int] = defaultdict(int)

    def __call__(self, entity: dict) -> str:
        key = entity["best_label"]
        self.counts[key] += 1
        return f"[{key}_{self.counts[key]}]"



from typing import Callable

from privateai_client.components import AnalyzeTextResponse

EntityProcessor = Callable[[dict], str]


def deidentify_text(
    text: list[str],
    response: AnalyzeTextResponse,
    entity_processors: dict[str, EntityProcessor],
    default_processor: EntityProcessor,
) -> list[str]:
    """
    Deidentifies analyzed text by processing entities with multiple processors, adjusting text dynamically.

    Args:
        text: The original list of text messages used as input in the `PAIClient.analyze_text()` call.
        response: The response object returned by `PAIClient.analyze_text()`, containing detected entities.
        entity_processors: A dictionary mapping entity types to processing functions in the format
            `{ENTITY_TYPE: entity_processor_fn}`, where `entity_processor_fn` takes an entity dictionary
            and returns modified text for all occurrences of that entity type.
        default_processor: A fallback function used to process entities not found in `entity_processors`.


    Returns:
        A list of de-identified text messages, with the same length and order as the `text` argument.

    """
    modified_texts = []
    for t, entities in zip(text, response.entities):
        offset = 0
        modified_text = t
        for entity in sorted(entities, key=lambda e: e["location"]["stt_idx"]):
            start_idx = entity["location"]["stt_idx"] + offset
            end_idx = entity["location"]["end_idx"] + offset

            processor = entity_processors.get(entity["best_label"], default_processor)
            modified_entity_text = processor(entity)

            length_diff = len(modified_entity_text) - len(entity["text"])
            offset += length_diff

            modified_text = (
                modified_text[:start_idx]
                + modified_entity_text
                + modified_text[end_idx:]
            )
        modified_texts.append(modified_text)
    return modified_texts



import json

import pytest
import requests

from ..components.pai_responses import AnalyzeTextResponse
from ..post_processing import (
    FuzzyMatchEntityProcessor,
    MarkerEntityProcessor,
    MaskEntityProcessor,
    deidentify_text,
)


# Deidentify text
def test_deidentify_text_no_entities():
    response = requests.Response()
    response.status_code = 200
    response._content = json.dumps(
        [
            {
                "entities": [],
                "entities_present": False,
                "characters_processed": 11,
            }
        ]
    ).encode("utf-8")

    text_out = deidentify_text(
        text=["My name is."],
        response=AnalyzeTextResponse(response),
        entity_processors={},
        default_processor=MarkerEntityProcessor(),
    )
    assert text_out == ["My name is."]


def test_deidentify_text_with_default_entities():
    response = requests.Response()
    response.status_code = 200
    response._content = json.dumps(
        [
            {
                "entities": [
                    {
                        "text": "John",
                        "location": {"stt_idx": 11, "end_idx": 15},
                        "best_label": "NAME_GIVEN",
                        "labels": {
                            "NAME": 0.9030465483665466,
                            "NAME_GIVEN": 0.8891170620918274,
                        },
                    }
                ],
                "entities_present": True,
                "characters_processed": 16,
            }
        ]
    ).encode("utf-8")

    text_out = deidentify_text(
        text=["My name is John."],
        response=AnalyzeTextResponse(response),
        entity_processors={},
        default_processor=MarkerEntityProcessor(),
    )
    assert text_out == ["My name is [NAME_GIVEN_1]."]


def test_deidentify_text_with_fuzzy_matching():
    response = requests.Response()
    response.status_code = 200
    response._content = json.dumps(
        [
            {
                "entities": [
                    {
                        "text": "John",
                        "location": {"stt_idx": 11, "end_idx": 15},
                        "best_label": "NAME_GIVEN",
                        "labels": {
                            "NAME": 0.9030465483665466,
                            "NAME_GIVEN": 0.8891170620918274,
                        },
                    }
                ],
                "entities_present": True,
                "characters_processed": 16,
            }
        ]
    ).encode("utf-8")

    text_out = deidentify_text(
        text=["My name is John."],
        response=AnalyzeTextResponse(response),
        entity_processors={
            "NAME_GIVEN": FuzzyMatchEntityProcessor(
                known_words_list=("Josh",),
                threshold=2,
                strategy="BLOCK",
                process_type="MASK",
            )
        },
        default_processor=MarkerEntityProcessor(),
    )
    assert text_out == ["My name is ####."]


# Mask processors
def test_mask_processor():
    entity_text = "John"
    entity = {
        "text": entity_text,
        "location": {"stt_idx": 11, "end_idx": 15},
        "best_label": "NAME_GIVEN",
    }
    processor = MaskEntityProcessor(masking_character="^")
    masked_entity = processor(entity)
    assert masked_entity == "^" * len(entity_text)


# Marker processor
def test_marker_processor():
    entity_1 = {
        "text": "John",
        "best_label": "NAME_GIVEN",
    }

    entity_2 = {
        "text": "France",
        "best_label": "LOCATION_COUNTRY",
    }

    entity_3 = {
        "text": "Josh",
        "best_label": "NAME_GIVEN",
    }
    processor = MarkerEntityProcessor()
    markers = [processor(e) for e in [entity_1, entity_2, entity_3]]
    assert markers == ["[NAME_GIVEN_1]", "[LOCATION_COUNTRY_1]", "[NAME_GIVEN_2]"]


# Fuzzy match processor
@pytest.mark.parametrize(
    argnames=[
        "entity_text",
        "known_words_list",
        "threshold",
        "strategy",
        "process_type",
        "masking_character",
        "ignore_casing",
        "processed_text",
    ],
    argvalues=[
        ("Josh", ["John", "Peter"], 2, "BLOCK", "MASK", "#", True, "####"),
        ("Ian", ["John", "Peter"], 2, "BLOCK", "MASK", "#", True, "Ian"),
        ("Josh", ["John", "Peter"], 2, "ALLOW", "MASK", "#", True, "Josh"),
        ("Ian", ["John", "Peter"], 2, "ALLOW", "MASK", "#", True, "###"),
        ("Josh", ["John", "Peter"], 2, "BLOCK", "MARKER", "#", True, "[NAME_GIVEN_1]"),
        ("Josh", ["JOSH", "Peter"], 2, "BLOCK", "MARKER", "#", False, "Josh"),
    ],
)
def test_fuzzy_match_entity_processor(
    entity_text,
    known_words_list,
    threshold,
    strategy,
    process_type,
    masking_character,
    ignore_casing,
    processed_text,
):
    processor = FuzzyMatchEntityProcessor(
        known_words_list=known_words_list,
        threshold=threshold,
        strategy=strategy,
        process_type=process_type,
        masking_character=masking_character,
        ignore_casing=ignore_casing,
    )
    entity = {"text": entity_text, "best_label": "NAME_GIVEN"}
    assert processor(entity) == processed_text


@pytest.mark.parametrize(
    argnames=[
        "known_words_list",
        "threshold",
        "strategy",
        "process_type",
        "masking_character",
        "ignore_casing",
        "error_str",
    ],
    argvalues=[
        (
            {"NAME": ["JOHN"]},
            2,
            "BLOCK",
            "MASK",
            "#",
            True,
            "Invalid value for known_words_list. Accepted are list, tuple or set of strings.",
        ),
        (
            ["John", 25],
            2,
            "BLOCK",
            "MASK",
            "#",
            True,
            "Invalid value for known_words_list. Accepted are list, tuple or set of strings.",
        ),
        (
            ["John", "Peter"],
            "two",
            "BLOCK",
            "MASK",
            "#",
            True,
            "Invalid value for threshold. Accepted value is a positive integer.",
        ),
        (
            ["John", "Peter"],
            0,
            "BLOCK",
            "MASK",
            "#",
            True,
            "Invalid value for threshold. Accepted value is a positive integer.",
        ),
        (
            ["John", "Peter"],
            2,
            "ENABLE",
            "MASK",
            "#",
            True,
            "Invalid value for strategy. Accepted values: 'BLOCK' and 'ALLOW'",
        ),
        (
            ["John", "Peter"],
            2,
            "ALLOW",
            "HIDE",
            "#",
            True,
            "Invalid value for process_type. Accepted values: 'MARKER' and 'MASK'",
        ),
        (
            ["John", "Peter"],
            2,
            "BLOCK",
            "MARKER",
            [],
            True,
            "Invalid value for masking_character. Accepted value is a valid string",
        ),
        (
            ["JOSH", "Peter"],
            2,
            "BLOCK",
            "MARKER",
            "#",
            "ignore",
            "Invalid value for ignore_casing. Accepted values: True and False",
        ),
    ],
)
def test_fuzzy_match_processor_invalid_attrs(
    known_words_list,
    threshold,
    strategy,
    process_type,
    masking_character,
    ignore_casing,
    error_str,
):
    with pytest.raises(ValueError, match=error_str):
        FuzzyMatchEntityProcessor(
            known_words_list=known_words_list,
            threshold=threshold,
            strategy=strategy,
            process_type=process_type,
            masking_character=masking_character,
            ignore_casing=ignore_casing,
        )
